package cc.fascinated.backend.service;

import cc.fascinated.backend.common.Timer;
import cc.fascinated.backend.leaderboard.impl.ScoreSaberLeaderboard;
import cc.fascinated.backend.model.account.Account;
import cc.fascinated.backend.model.token.ScoreSaberAccountToken;
import cc.fascinated.backend.repository.AccountRepository;
import cc.fascinated.backend.repository.ScoreRepository;
import lombok.extern.log4j.Log4j2;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.util.List;
import java.util.Optional;
import java.util.concurrent.TimeUnit;

@Service @Log4j2(topic = "Account Service")
public class AccountService {

    /**
     * How often the account should be updated.
     */
    private static final long UPDATE_INTERVAL = TimeUnit.HOURS.toMillis(1);

    /**
     * The {@link AccountRepository} instance.
     */
    private final AccountRepository accountRepository;

    /**
     * The {@link ScoreRepository} instance.
     */
    private final ScoreRepository scoreRepository;

    /**
     * The {@link ScoreSaberService} instance.
     */
    private final ScoreSaberService scoreSaberService;

    @Autowired
    public AccountService(AccountRepository accountRepository, ScoreRepository scoreRepository, ScoreSaberService scoreSaberService) {
        this.accountRepository = accountRepository;
        this.scoreRepository = scoreRepository;
        this.scoreSaberService = scoreSaberService;

        Timer.scheduleRepeating(() -> {
            List<Account> accounts = accountRepository.findAll();
            log.info("Updating accounts.");
            for (Account account : accounts) {
                updateAccount(account);
            }
            log.info("Updated {} accounts.", accounts.size());
        }, 0, UPDATE_INTERVAL);
    }

    /**
     * Gets the ScoreSaber account.
     * <p>
     *     If the account is not found in the database,
     *     it will be fetched from the ScoreSaber API,
     *     fetch all the scores for the account then
     *     save the account to the database.
     * </p>
     *
     * @param id The id of the account.
     * @return The account.
     */
    public Account getAccount(String id) {
        log.info("Fetching account '{}'.", id);
        Optional<Account> optionalAccount = accountRepository.findById(id);
        if (optionalAccount.isEmpty()) {
            log.info("Account '{}' not found in the database. Fetching from ScoreSaber API.", id);
            return updateAccount(Account.fromToken(scoreSaberService.getAccount(id)));
        }
        log.info("Account '{}' found in the database.", id);
        return optionalAccount.get();
    }

    /**
     * Fetches the account from the ScoreSaber API
     * and saves it to the database.
     *
     * @param account The account.
     */
    public Account updateAccount(Account account) {
        String id = account.getId();

        // Fetch the account from the ScoreSaber API.
        ScoreSaberAccountToken accountToken = scoreSaberService.getAccount(id); // Fetch the account from the ScoreSaber API.
        if (accountToken == null) {
            log.warn("Account '{}' not found in the ScoreSaber API.", id);
            return null;
        }

        // Update the account with the new token.
        account = Account.fromToken(accountToken);
        accountRepository.save(account);

        // Don't fetch new scores if the account is inactive or banned.
        if (account.isInactive() || account.isBanned()) {
            return account;
        }

        // Fetch the scores for the account.
        scoreSaberService.updateScores(account);

        // Set the raw pp per +1 global pp
        double rawPerGlobalPP = ScoreSaberLeaderboard.INSTANCE.getRawPerGlobalPP(scoreRepository.getRankedScores(id), 1);
        account.setRawPerGlobalPerformancePoints(rawPerGlobalPP);
        return accountRepository.save(account);
    }
}
